"""Logging Module.

Module which setsup the basic logging infrustrcuture for the application.
"""

import logging
import sys

# logger formating
BRIEF_FORMAT = "%(levelname)s %(asctime)s - %(name)s: %(message)s"
VERBOSE_FORMAT = (
    "%(levelname)s|%(asctime)s|%(name)s|%(filename)s|"
    "%(funcName)s|%(lineno)d: %(message)s"
)
FORMAT_TO_USE = VERBOSE_FORMAT

# logger levels
DEBUG = logging.DEBUG
INFO = logging.INFO
WARN = logging.WARN
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL


def get_logger(name=None, log_level=logging.INFO):
    """Set the basic logging features for the application.

    Parameters
    ----------
    name : str, optional
        The name of the logger. Defaults to ``None``
    log_level : int, optional
        The logging level. Defaults to ``logging.INFO``

    Returns
    -------
    logging.Logger
        Returns a Logger obejct which is set with the passed in paramters.
    """
    logging.basicConfig(format=FORMAT_TO_USE, stream=sys.stdout, level=log_level)
    logger = logging.getLogger(name)
    return logger



from unittest.mock import patch

import pytest

from edgetest.core import TestPackage
from edgetest.report import gen_report


@patch("edgetest.report.tabulate", autospec=True)
@patch("edgetest.core.TestPackage.upgraded_packages", autospec=True)
@patch("edgetest.core.TestPackage.lowered_packages", autospec=True)
def test_report(
    mock_test_package_lowered, mock_test_package, mock_tabulate, plugin_manager
):
    """Test gen_report function"""

    tester_list = [
        TestPackage(hook=plugin_manager.hook, envname="myenv", upgrade=["myupgrade1"]),
        TestPackage(hook=plugin_manager.hook, envname="myenv", upgrade=["myupgrade2"]),
        TestPackage(hook=plugin_manager.hook, envname="myenv_lower", lower=["mylower"]),
    ]

    expected_headers = [
        "Environment",
        "Setup successful",
        "Passing tests",
        "Upgraded packages",
        "Lowered packages",
        "Package version",
    ]
    gen_report(tester_list)
    mock_tabulate.assert_called_with([], headers=expected_headers, tablefmt="rst")

    gen_report(tester_list, output_type="github")
    mock_tabulate.assert_called_with([], expected_headers, tablefmt="github")

    with pytest.raises(ValueError):
        gen_report(tester_list, output_type="bad")



"""Set up a fake hook for testing purposes."""

import platform
from pathlib import Path
from typing import Dict, List

import pluggy
import pytest

from edgetest import hookspecs

hookimpl = pluggy.HookimplMarker("edgetest")


class FakeHook:
    """Create a series of fake hooks."""

    @hookimpl
    def path_to_python(self, basedir: str, envname: str) -> str:
        """Return the path to the python executable."""
        # return str(Path(basedir, envname, "bin", "python"))
        if platform.system() == "Windows":
            return str(Path(basedir) / envname / "Scripts" / "python")
        else:
            return str(Path(basedir) / envname / "bin" / "python")

    @hookimpl
    def create_environment(self, basedir: str, envname: str, conf: Dict):
        """Create the virtual environment for testing."""
        pass

    @hookimpl
    def run_update(self, basedir: str, envname: str, upgrade: List):
        """Update packages from upgrade list."""
        pass

    @hookimpl
    def run_install_lower(basedir: str, envname: str, lower: Dict[str, str]):
        """Install lower bounds of packages from lower list."""
        pass


class FakeHookEnvironmentError:
    """Create a series of fake hooks that raise errors."""

    @hookimpl
    def path_to_python(self, basedir: str, envname: str) -> str:
        """Return the path to the python executable."""
        # return str(Path(basedir, envname, "bin", "python"))
        if platform.system() == "Windows":
            return str(Path(basedir) / envname / "Scripts" / "python")
        else:
            return str(Path(basedir) / envname / "bin" / "python")

    @hookimpl
    def create_environment(self, basedir: str, envname: str, conf: Dict):
        """Create the virtual environment for testing."""
        raise RuntimeError("Could not create virtual environment.")

    @hookimpl
    def run_update(self, basedir: str, envname: str, upgrade: List):
        """Update packages from upgrade list."""
        pass

    @hookimpl
    def run_install_lower(basedir: str, envname: str, lower: Dict[str, str]):
        """Install lower bounds of packages from lower list."""
        pass


class FakeHookUpdateError:
    """Create a series of fake hooks that raise errors."""

    @hookimpl
    def path_to_python(self, basedir: str, envname: str) -> str:
        """Return the path to the python executable."""
        # return str(Path(basedir, envname, "bin", "python"))
        if platform.system() == "Windows":
            return str(Path(basedir) / envname / "Scripts" / "python")
        else:
            return str(Path(basedir) / envname / "bin" / "python")

    @hookimpl
    def create_environment(self, basedir: str, envname: str, conf: Dict):
        """Create the virtual environment for testing."""
        pass

    @hookimpl
    def run_update(self, basedir: str, envname: str, upgrade: List):
        """Update packages from upgrade list."""
        raise RuntimeError("Could not update packages.")

    @hookimpl
    def run_install_lower(basedir: str, envname: str, lower: Dict[str, str]):
        """Install lower bounds of packages from lower list."""
        pass


class FakeHookLowerError:
    """Create a series of fake hooks that raise errors."""

    @hookimpl
    def path_to_python(self, basedir: str, envname: str) -> str:
        """Return the path to the python executable."""
        # return str(Path(basedir, envname, "bin", "python"))
        if platform.system() == "Windows":
            return str(Path(basedir) / envname / "Scripts" / "python")
        else:
            return str(Path(basedir) / envname / "bin" / "python")

    @hookimpl
    def create_environment(self, basedir: str, envname: str, conf: Dict):
        """Create the virtual environment for testing."""
        pass

    @hookimpl
    def run_update(self, basedir: str, envname: str, upgrade: List):
        """Update packages from upgrade list."""
        pass

    @hookimpl
    def run_install_lower(basedir: str, envname: str, lower: Dict[str, str]):
        """Install lower bounds of packages from lower list."""
        raise RuntimeError("Could not install lower bounds of packages.")



@pytest.fixture
def plugin_manager():
    """The plugin manager for our fake hook."""
    pm = pluggy.PluginManager("edgetest")
    pm.add_hookspecs(hookspecs)
    pm.register(FakeHook())

    return pm

@pytest.fixture
def plugin_manager_environment_error():
    """The plugin manager for our fake hook that raises errors."""
    pm = pluggy.PluginManager("edgetest")
    pm.add_hookspecs(hookspecs)
    pm.register(FakeHookEnvironmentError())

    return pm

@pytest.fixture
def plugin_manager_upgrade_error():
    """The plugin manager for our fake hook that raises errors."""
    pm = pluggy.PluginManager("edgetest")
    pm.add_hookspecs(hookspecs)
    pm.register(FakeHookUpdateError())

    return pm

@pytest.fixture
def plugin_manager_lower_error():
    """The plugin manager for our fake hook that raises errors."""
    pm = pluggy.PluginManager("edgetest")
    pm.add_hookspecs(hookspecs)
    pm.register(FakeHookLowerError())

    return pm


"""Default virtual environment hook."""

import platform
from pathlib import Path
from typing import Dict, List
from venv import EnvBuilder

import pluggy

from .utils import _run_command

hookimpl = pluggy.HookimplMarker("edgetest")


@hookimpl(trylast=True)
def path_to_python(basedir: str, envname: str) -> str:
    """Return the path to the python executable."""
    if platform.system() == "Windows":
        return str(Path(basedir) / envname / "Scripts" / "python")
    else:
        return str(Path(basedir) / envname / "bin" / "python")


@hookimpl(trylast=True)
def create_environment(basedir: str, envname: str, conf: Dict):
    """Create the virtual environment for testing.

    Creates an environment using ``venv``.

    Parameters
    ----------
    basedir : str
        The base directory location for the environment.
    envname : str
        The name of the virtual environment.
    conf : dict
        Ignored.

    Raises
    ------
    RuntimeError
        Error raised if the environment cannot be created.
    """
    builder = EnvBuilder(with_pip=True)
    try:
        builder.create(env_dir=Path(basedir, envname))
    except Exception:
        raise RuntimeError(f"Unable to create {envname} in {basedir}")


@hookimpl(trylast=True)
def run_update(basedir: str, envname: str, upgrade: List, conf: Dict):
    """Update packages from upgrade list.

    Parameters
    ----------
    basedir : str
        The base directory location for the environment.
    envname : str
        The name of the virtual environment.
    upgrade : list
        The list of packages to upgrade
    conf : dict
        Ignored.

    Raises
    ------
    RuntimeError
        Error raised if the packages cannot be updated.
    """
    python_path = path_to_python(basedir, envname)
    try:
        _run_command(
            python_path,
            "-m",
            "pip",
            "install",
            *upgrade,
            "--upgrade",
        )
    except Exception:
        raise RuntimeError(f"Unable to pip upgrade: {upgrade}")


@hookimpl(trylast=True)
def run_install_lower(basedir: str, envname: str, lower: List[str], conf: Dict):
    """Install lower bounds of packages provided.

    Parameters
    ----------
    basedir : str
        The base directory location for the environment.
    envname : str
        Environment to install into.
    lower : List[str]
        Lower bounds of packages to install.
    conf : Dict
        The configuration dictionary for the environment. This is useful if you
        want to add configuration arguments for additional dependencies that can
        only be installed through the environment manager (e.g. Conda).
    """
    python_path = path_to_python(basedir, envname)
    try:
        _run_command(
            python_path,
            "-m",
            "pip",
            "install",
            *lower,
        )
    except Exception:
        raise RuntimeError(f"Unable to pip install: {lower}")



"""Utility functions."""

import os
from configparser import ConfigParser
from contextlib import contextmanager
from copy import deepcopy
from pathlib import Path
from subprocess import PIPE, Popen
from typing import Any, Dict, List, Optional, Tuple, Union

from packaging.specifiers import SpecifierSet
from pkg_resources import parse_requirements
from tomlkit import TOMLDocument, load
from tomlkit.container import Container
from tomlkit.items import Array, Item, String, Table

from .logger import get_logger

LOG = get_logger(__name__)


def _run_command(*args) -> Tuple[str, int]:
    """Run a command using ``subprocess.Popen``.

    Parameters
    ----------
    *args
        Arguments for the command.

    Returns
    -------
    str
        The output
    int
        The exit code

    Raises
    ------
    RuntimeError
        Error raised when the command is not successfully executed.
    """
    LOG.debug(f"Running the following command: \n\n {' '.join(args)}")
    popen = Popen(args, stdout=PIPE, universal_newlines=True)
    out, _ = popen.communicate()
    if popen.returncode:
        raise RuntimeError(
            f"Unable to run the following command: \n\n {' '.join(args)}"
        )

    return out, popen.returncode


@contextmanager
def pushd(new_dir: str):
    """Create a context manager for running commands in sub-directories.

    Parameters
    ----------
    new_dir : str
        The relative directory to run the command in.
    """
    curr_dir = Path.cwd()
    os.chdir(curr_dir / new_dir)
    try:
        yield
    finally:
        os.chdir(curr_dir)


def _convert_toml_array_to_string(item: Union[Item, Any]) -> str:
    if isinstance(item, Array):
        return "\n".join(item)
    elif isinstance(item, String):
        return str(item)
    else:
        raise ValueError


def convert_requirements(requirements: str, conf: Optional[Dict] = None) -> Dict:
    """Generate environments for a newline-separate list of package requirements.

    This function will generate one environment per entry with an additional environment
    that upgrades all requirements simultaneously.

    Parameters
    ----------
    requirements : str
        The requirements string.
    conf : dict, optional (default None)
        An existing configuration to edit.

    Returns
    -------
    Dict
        A configuration dictionary.
    """
    conf = {"envs": []} if conf is None else conf
    pkgs = [pkg.project_name for pkg in parse_requirements(requirements)]
    for pkg in pkgs:
        conf["envs"].append({})
        conf["envs"][-1]["name"] = pkg
        conf["envs"][-1]["upgrade"] = pkg
    # Create an environment with all requirements upgraded
    conf["envs"].append({})
    conf["envs"][-1]["name"] = "all-requirements"
    conf["envs"][-1]["upgrade"] = "\n".join(pkgs)

    return conf


def gen_requirements_config(fname_or_buf: str, **options) -> Dict:
    """Generate a configuration file from package requirements.

    This function will convert the package installation requirements to a configuration
    file with one environment per requirement.

    Parameters
    ----------
    fname_or_buf : str
        Path to the requirements file to parse using ``pkg_resources.parse_requirements``
        or the string representing the requirements file.
    **options
        Options to apply to each test environment.

    Returns
    -------
    Dict
        The configuration file.
    """
    # First, get the requirements
    if Path(fname_or_buf).is_file():
        with open(fname_or_buf) as infile:
            cfg = infile.read()
    else:
        cfg = fname_or_buf
    output = convert_requirements(requirements=cfg)
    for index in range(len(output["envs"])):
        output["envs"][index].update(options)

    return output


def parse_cfg(filename: str = "setup.cfg", requirements: Optional[str] = None) -> Dict:
    """Generate a configuration from a ``.ini`` style file.

    This function can operate in two ways. First, it can look for sections that
    start with ``edgetest`` and build a configuration. Suppose
    you have ``setup.cfg`` as follows:

    .. code-block:: ini

        [edgetest.envs.pandas]
        upgrade =
            pandas

    This will result in a configuration that has one testing environment, named
    ``pandas``, that upgrades the ``pandas`` package.

    If you don't have any sections that start with ``edgetest.envs``, we will look for
    the PEP 517-style ``setup.cfg`` install requirements (the ``install_requires`` key
    within the ``options`` section). To set global defaults for you environments, use
    the ``edgetest`` section:

    .. code-block:: ini

        [edgetest]
        extras =
            tests
        command =
            pytest tests -m "not integration"

        [edgetest.envs.pandas]
        upgrade =
            pandas

    For this single environment file, the above configuration is equivalent to

    .. code-block:: ini

        [edgetest.envs.pandas]
        extras =
            tests
        command =
            pytest tests -m "not integration"
        upgrade =
            pandas

    Parameters
    ----------
    filename : str, optional (default "setup.cfg")
        The name of the configuration file to read. Defaults to ``setup.cfg``.
    requirements : str, optional (default None)
        An optional path to the requirements text file. If there are no PEP-517
        style dependencies or coded environments in the edgetest configuration, this
        function will look for dependencies in the requirements file.

    Returns
    -------
    Dict
        A configuration dictionary for ``edgetest``.
    """
    # Read in the configuration file
    config = ConfigParser()
    config.read(filename)
    # Parse
    output: Dict = {"envs": []}
    # Get any global options if necessary
    options = dict(config["edgetest"]) if "edgetest" in config else {}
    # Next, create the sections
    for section in config.sections():
        if not section.startswith("edgetest."):
            continue
        # Look for the special ``envs`` key
        section_name = section.split(".")
        if section_name[1] == "envs":
            output["envs"].append(dict(config[section]))
            output["envs"][-1]["name"] = section_name[2]
            if (
                "lower" in output["envs"][-1]
                and "options" in config
                and "install_requires" in config["options"]
            ):
                output["envs"][-1]["lower"] = get_lower_bounds(
                    config.get("options", "install_requires"),
                    output["envs"][-1]["lower"],
                )
        else:
            output[section_name[1]] = dict(config[section])
    if len(output["envs"]) == 0:
        if config.get("options", "install_requires"):
            output = convert_requirements(
                requirements=config["options"]["install_requires"], conf=output
            )
        elif requirements:
            req_conf = gen_requirements_config(fname_or_buf=requirements)
            output["envs"] = req_conf["envs"]
        else:
            raise ValueError("Please supply a valid list of environments to create.")
    # Apply global environment options (without overwriting)
    for idx in range(len(output["envs"])):
        output["envs"][idx] = dict(
            list(options.items()) + list(output["envs"][idx].items())
        )

    return output


def parse_toml(
    filename: str = "pyproject.toml", requirements: Optional[str] = None
) -> Dict:
    """Generate a configuration from a ``.toml`` style file.

    This function can operate in two ways. First, it will look for tables that
    start with ``edgetest`` and build a configuration. Suppose
    you have ``pyproject.toml`` as follows:

    .. code-block:: toml

        [edgetest.envs.pandas]
        upgrade = [
            "pandas"
        ]

    This will result in a configuration that has one testing environment, named
    ``pandas``, that upgrades the ``pandas`` package.

    If you don't have any tables that start with ``edgetest.envs``, we will look for
    the installation requirements (the ``dependencies`` key within the ``project`` section).
    To set global defaults for you environments, use the ``edgetest`` table:

    .. code-block:: toml

        [edgetest]
        extras = [
            "tests"
        ]
        command = "pytest tests -m 'not integration'"

        [edgetest.envs.pandas]
        upgrade = [
            "pandas"
        ]

    For this single environment file, the above configuration is equivalent to

    .. code-block:: toml

        [edgetest.envs.pandas]
        extras = [
            "tests"
        ]
        command = "pytest tests -m 'not integration'"
        upgrade = [
            "pandas"
        ]

    Parameters
    ----------
    filename : str, optional (default "pyproject.toml")
        The name of the toml file to read. Defaults to ``pyproject.toml``.
    requirements : str, optional (default None)
        An optional path to the requirements text file. If there are no TOML
        style dependencies or coded environments in the edgetest configuration, this
        function will look for dependencies in the requirements file.

    Returns
    -------
    Dict
        A configuration dictionary for ``edgetest``.
    """
    options: Union[Item, Container, dict]
    # Read in the configuration file
    config: TOMLDocument = load(open(filename))
    # Parse
    output: Dict = {"envs": []}
    # Get any global options if necessary. First scan through and pop out any Tables
    temp_config = deepcopy(config)
    if "edgetest" in config:
        for j in config["edgetest"].items():  # type: ignore
            if isinstance(config["edgetest"][j[0]], Table):  # type: ignore
                _ = temp_config["edgetest"].pop(  # type: ignore
                    j[0], None
                )  # remove Tables from the temp config
            else:
                temp_config["edgetest"][j[0]] = _convert_toml_array_to_string(  # type: ignore
                    temp_config["edgetest"][j[0]]  # type: ignore
                )
        options = temp_config["edgetest"]
    else:
        options = {}

    # Check envs exists and any other Tables
    if "edgetest" in config:
        for section in config["edgetest"]:  # type: ignore
            if section == "envs":
                for env in config["edgetest"]["envs"]:  # type: ignore
                    for item in config["edgetest"]["envs"][env]:  # type: ignore
                        # If an Array then decompose to a string format
                        config["edgetest"]["envs"][env][  # type: ignore
                            item
                        ] = _convert_toml_array_to_string(
                            config["edgetest"]["envs"][env][item]  # type: ignore
                        )
                    output["envs"].append(dict(config["edgetest"]["envs"][env]))  # type: ignore
                    output["envs"][-1]["name"] = env
                    if (
                        "lower" in output["envs"][-1]
                        and "project" in config
                        and "dependencies" in config["project"]  # type: ignore
                    ):
                        output["envs"][-1]["lower"] = get_lower_bounds(
                            dict(config["project"])["dependencies"],  # type: ignore
                            output["envs"][-1]["lower"],
                        )
            elif isinstance(config["edgetest"][section], Table):  # type: ignore
                output[section] = dict(config["edgetest"][section])  # type: ignore

    if len(output["envs"]) == 0:
        if config.get("project").get("dependencies"):  # type: ignore
            output = convert_requirements(
                requirements="\n".join(config["project"]["dependencies"]),  # type: ignore
                conf=output,  # type: ignore # noqa: E501
            )
        elif requirements:
            req_conf = gen_requirements_config(fname_or_buf=requirements)
            output["envs"] = req_conf["envs"]
        else:
            raise ValueError("Please supply a valid list of environments to create.")

    # Apply global environment options (without overwriting)
    for idx in range(len(output["envs"])):
        output["envs"][idx] = dict(
            list(options.items()) + list(output["envs"][idx].items())  # type: ignore
        )

    return output


def upgrade_requirements(
    fname_or_buf: str, upgraded_packages: List[Dict[str, str]]
) -> str:
    """Create an upgraded requirements file.

    Parameters
    ----------
    fname_or_buf : str
        Path to the requirements file to parse using ``pkg_resources.parse_requirements``
        or the string representing the requirements file.
    upgraded_packages : list
        A list of packages upgraded in the testing procedure.

    Returns
    -------
    str
        The string file representing the new requirements file.
    """
    # Get the existing file
    try:
        if Path(fname_or_buf).is_file():
            with open(fname_or_buf) as infile:
                cfg = infile.read()
        else:
            cfg = fname_or_buf
    except OSError:
        # Filename too long for the is_file() function
        cfg = fname_or_buf
    pkgs = [pkg for pkg in parse_requirements(cfg)]
    upgrades = {pkg["name"]: pkg["version"] for pkg in upgraded_packages}

    for pkg in pkgs:
        if pkg.project_name not in upgrades:
            continue
        # Replace the spec
        specs = deepcopy(pkg.specs)
        for index, value in enumerate(specs):
            if value[0] == "<=":
                pkg.specs[index] = ("<=", upgrades[pkg.project_name])
            elif value[0] == "<":
                pkg.specs[index] = ("!=", value[1])
                pkg.specs.append(("<=", upgrades[pkg.project_name]))
            elif value[0] == "==":
                pkg.specs = [(">=", value[1]), ("<=", upgrades[pkg.project_name])]
        pkg.specifier = SpecifierSet(",".join("".join(spec) for spec in pkg.specs))  # type: ignore

    return "\n".join(str(pkg) for pkg in pkgs)


def upgrade_setup_cfg(
    upgraded_packages: List[Dict[str, str]], filename: str = "setup.cfg"
) -> ConfigParser:
    """Upgrade the ``setup.cfg`` file.

    Parameters
    ----------
    upgraded_packages : List[Dict[str, str]]
        A list of packages upgraded in the testing procedure.
    filename : str, optional (default "setup.cfg")
        The name of the configuration file to read. Defaults to ``setup.cfg``.

    Returns
    -------
    ConfigParser
        The updated configuration file.
    """
    parser = ConfigParser()
    parser.read(filename)
    if "options" in parser and parser.get("options", "install_requires"):
        LOG.info(f"Updating the requirements in {filename}")
        upgraded = upgrade_requirements(
            fname_or_buf=parser["options"]["install_requires"].lstrip(),
            upgraded_packages=upgraded_packages,
        )
        parser["options"]["install_requires"] = "\n" + upgraded
    # Update the extras, if necessary
    if "options.extras_require" in parser:
        for extra, dependencies in parser.items("options.extras_require"):
            upgraded = upgrade_requirements(
                fname_or_buf=dependencies,
                upgraded_packages=upgraded_packages,
            )
            parser["options.extras_require"][extra] = "\n" + upgraded

    return parser


def upgrade_pyproject_toml(
    upgraded_packages: List[Dict[str, str]], filename: str = "pyproject.toml"
) -> TOMLDocument:
    """Upgrade the ``pyproject.toml`` file.

    Parameters
    ----------
    upgraded_packages : List[Dict[str, str]]
        A list of packages upgraded in the testing procedure.
    filename : str, optional (default "pyproject.toml")
        The name of the configuration file to read. Defaults to ``pyproject.toml``.

    Returns
    -------
    TOMLDocument
        The updated TOMLDocument.
    """
    parser: TOMLDocument = load(open(filename))
    if "project" in parser and parser.get("project").get("dependencies"):  # type: ignore
        LOG.info(f"Updating the requirements in {filename}")
        upgraded = upgrade_requirements(
            fname_or_buf="\n".join(parser["project"]["dependencies"]),  # type: ignore
            upgraded_packages=upgraded_packages,
        )
        parser["project"]["dependencies"] = upgraded.split("\n")  # type: ignore
    # Update the extras, if necessary
    if parser.get("project").get("optional-dependencies"):  # type: ignore
        for extra, dependencies in parser["project"]["optional-dependencies"].items():  # type: ignore # noqa: E501
            upgraded = upgrade_requirements(
                fname_or_buf="\n".join(dependencies),
                upgraded_packages=upgraded_packages,
            )
            parser["project"]["optional-dependencies"][extra] = upgraded.split("\n")  # type: ignore

    return parser


def _isin_case_dashhyphen_ins(a: str, vals: List[str]) -> bool:
    """Run isin check that is case and dash/hyphen insensitive.

    Paramaters
    ----------
    a : str
        String value to check for membership against ``vals``.
    vals : list of str
        List of strings to check ``a`` against.

    Returns
    -------
    bool
        Return ``True`` if ``a`` in vals, otherwise ``False``.
    """
    for b in vals:
        if a.replace("_", "-").lower() == b.replace("_", "-").lower():
            return True
    return False


def get_lower_bounds(requirements: str, lower: str) -> str:
    r"""Get lower bounds of requested packages from installation requirements.

    Parses through the project ``requirements`` and the newline-delimited
    packages requested in ``lower`` to find the lower bounds.

    Parameters
    ----------
    requirements : str
        Project setup requirements,
        e.g. ``"pandas>=1.5.1,<=1.4.2\nnumpy>=1.22.1,<=1.25.4"``
    lower : str
        Newline-delimited packages requested,
         e.g. ``"pandas\nnumpy"``.

    Returns
    -------
    str
        The packages along with the lower bound, e.g. ``"pandas==1.5.1\nnumpy==1.22.1"``.
    """
    all_lower_bounds = {
        pkg.project_name + (f"[{','.join(pkg.extras)}]" if pkg.extras else ""): dict(
            pkg.specs
        ).get(">=")
        for pkg in parse_requirements(requirements)
    }

    lower_with_bounds = ""
    for pkg_name, lower_bound in all_lower_bounds.items():
        # TODO: Parse through extra requirements as well to get lower bounds
        if lower_bound is None:
            LOG.warning(
                "Requested %s lower bound, but did not find in install requirements.",
                pkg_name,
            )
        elif _isin_case_dashhyphen_ins(pkg_name, lower.split("\n")):
            lower_with_bounds += f"{pkg_name}=={lower_bound}\n"

    return lower_with_bounds



"""Test utility functions."""

from pathlib import Path
from unittest.mock import mock_open, patch

from tomlkit import array, string

from edgetest.schema import BASE_SCHEMA, EdgetestValidator, Schema
from edgetest.utils import (
    _convert_toml_array_to_string,
    _isin_case_dashhyphen_ins,
    gen_requirements_config,
    get_lower_bounds,
    parse_cfg,
    parse_toml,
    upgrade_pyproject_toml,
    upgrade_setup_cfg,
)

REQS = """
mydep1>=0.1.0,<=0.2.0
mydep2<=0.10.0
"""

CFG_NOREQS = """
[edgetest.envs.myenv]
upgrade =
    myupgrade
command =
    pytest tests -m 'not integration'

[edgetest.envs.myenv_lower]
lower =
    mylower
command =
    pytest tests -m 'not integration'
"""

CFG_REQS = """
[options]
install_requires =
    myupgrade
"""

CFG_REQS_DEFAULTS = """
[options]
install_requires =
    myupgrade

[edgetest]
extras =
    tests
command =
    pytest tests -m 'not integration'
"""

CFG_NOREQS_DEFAULTS = """
[edgetest]
extras =
    tests
command =
    pytest tests -m 'not integration'

[edgetest.envs.myenv]
upgrade =
    myupgrade
command =
    pytest tests

[edgetest.envs.myenv_lower]
lower =
    mylower
command =
    pytest tests
"""

CFG_CUSTOM = """
[edgetest]
extras =
    tests
command =
    pytest tests -m 'not integration'

[edgetest.custom]
mycustom = mykey

[edgetest.envs.myenv]
upgrade =
    myupgrade

[edgetest.envs.myenv_lower]
lower =
    mylower
"""


TOML_NOREQS = """
[edgetest.envs.myenv]
upgrade = [
    "myupgrade"
]
command = "pytest tests -m 'not integration'"

[edgetest.envs.myenv_lower]
lower = ["mylower"]
command = "pytest tests -m 'not integration'"
"""


TOML_REQS = """
[project]
dependencies = [
    "myupgrade"
]
"""

TOML_REQS_DEFAULTS = """
[project]
dependencies = [
    "myupgrade"
]
[edgetest]
extras = ["tests"]
command = "pytest tests -m 'not integration'"
"""

TOML_NOREQS_DEFAULTS = """
[edgetest]
extras = ["tests"]
command = "pytest tests -m 'not integration'"

[edgetest.envs.myenv]
upgrade = ["myupgrade"]
command = "pytest tests"

[edgetest.envs.myenv_lower]
lower = ["mylower"]
command = "pytest tests"
"""

TOML_CUSTOM = """
[edgetest]
extras = ["tests"]
command = "pytest tests -m 'not integration'"

[edgetest.custom]
mycustom = "mykey"

[edgetest.envs.myenv]
upgrade = ["myupgrade"]

[edgetest.envs.myenv_lower]
lower = ["mylower"]
"""


CFG_REQS_UPGRADE = """
[options]
install_requires =
    pandas<=1.0.0,>=1.0.0
    numpy<=1.0.0,>=1.0.0

[options.extras_require]
tests =
    pytest<=1.0.0,>=1.0.0
"""

TOML_REQS_UPGRADE = """
[project]
dependencies = [
    "pandas<=1.0.0,>=1.0.0",
    "numpy<=1.0.0,>=1.0.0",
]
[project.optional-dependencies]
tests = [
    "pytest<=1.0.0,>=1.0.0",
]
"""

REQS_NORMAL = """
pandas>=1.0.0,<=2.0
numpy<=0.24,>=0.01
"""


def test_get_lower_bounds():
    """Test getting lower bound from reqs."""
    assert get_lower_bounds(REQS_NORMAL, "pandas\nnumpy\n") == "pandas==1.0.0\nnumpy==0.01\n"
    assert get_lower_bounds(REQS_NORMAL, "pandas") == "pandas==1.0.0\n"
    assert get_lower_bounds(REQS_NORMAL, "") == ""
    assert get_lower_bounds(REQS, "mydep2") == ""

@patch("edgetest.utils.Path")
def test_parse_reqs(mock_pathlib):
    """Test creating a configuration from requirements."""
    mock_pathlib.return_value.is_file.return_value = True
    with patch("edgetest.utils.open", mock_open(read_data=REQS)):
        cfg = gen_requirements_config("filename")

    assert cfg == {
        "envs": [
            {"name": "mydep1", "upgrade": "mydep1"},
            {"name": "mydep2", "upgrade": "mydep2"},
            {"name": "all-requirements", "upgrade": "mydep1\nmydep2"},
        ]
    }
    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(cfg)


def test_parse_cfg(tmpdir):
    """Test parsing a config with no install requirements."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "myconfig.ini")
    with open(conf_loc, "w") as outfile:
        outfile.write(CFG_NOREQS)

    cfg = parse_cfg(filename=conf_loc)

    assert cfg == {
        "envs": [
            {
                "name": "myenv",
                "upgrade": "\nmyupgrade",
                "command": "\npytest tests -m 'not integration'",
            },
            {
                "name": "myenv_lower",
                "lower": "\nmylower",
                "command": "\npytest tests -m 'not integration'",
            },
        ]
    }

    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(cfg)


def test_parse_cfg_default(tmpdir):
    """Test parsing a config with no install requirements and defaults."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "myconfig.ini")
    with open(conf_loc, "w") as outfile:
        outfile.write(CFG_NOREQS_DEFAULTS)

    cfg = parse_cfg(filename=conf_loc)

    assert cfg == {
        "envs": [
            {
                "name": "myenv",
                "upgrade": "\nmyupgrade",
                "extras": "\ntests",
                "command": "\npytest tests",
            },
            {
                "name": "myenv_lower",
                "lower": "\nmylower",
                "extras": "\ntests",
                "command": "\npytest tests",
            },
        ]
    }

    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(cfg)


def test_parse_cfg_reqs(tmpdir):
    """Test parsing a PEP-517 style config."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "setup.cfg")
    with open(conf_loc, "w") as outfile:
        outfile.write(CFG_REQS)

    cfg = parse_cfg(filename=conf_loc)

    assert cfg == {
        "envs": [
            {"name": "myupgrade", "upgrade": "myupgrade"},
            {"name": "all-requirements", "upgrade": "myupgrade"},
        ]
    }

    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(cfg)


def test_parse_cfg_reqs_default(tmpdir):
    """Test parsing a PEP-517 style config with default arguments."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "setup.cfg")
    with open(conf_loc, "w") as outfile:
        outfile.write(CFG_REQS_DEFAULTS)

    cfg = parse_cfg(filename=conf_loc)

    assert cfg == {
        "envs": [
            {
                "name": "myupgrade",
                "upgrade": "myupgrade",
                "extras": "\ntests",
                "command": "\npytest tests -m 'not integration'",
            },
            {
                "name": "all-requirements",
                "upgrade": "myupgrade",
                "extras": "\ntests",
                "command": "\npytest tests -m 'not integration'",
            },
        ]
    }

    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(cfg)


def test_parse_custom_cfg(tmpdir):
    """Test parsing a custom configuration."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "setup.cfg")
    with open(conf_loc, "w") as outfile:
        outfile.write(CFG_CUSTOM)

    cfg = parse_cfg(filename=conf_loc)

    assert cfg == {
        "custom": {"mycustom": "mykey"},
        "envs": [
            {
                "name": "myenv",
                "upgrade": "\nmyupgrade",
                "extras": "\ntests",
                "command": "\npytest tests -m 'not integration'",
            },
            {
                "name": "myenv_lower",
                "lower": "\nmylower",
                "extras": "\ntests",
                "command": "\npytest tests -m 'not integration'",
            },
        ],
    }

    schema = Schema()
    schema.add_globaloption(
        "custom", {"type": "dict", "schema": {"mycustom": {"type": "string"}}}
    )

    validator = EdgetestValidator(schema=schema.schema)

    assert validator.validate(cfg)


def test_parse_toml(tmpdir):
    """Test parsing a config with no install requirements."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "myconfig.toml")
    with open(conf_loc, "w") as outfile:
        outfile.write(TOML_NOREQS)

    toml = parse_toml(filename=conf_loc)

    assert toml == {
        "envs": [
            {
                "name": "myenv",
                "upgrade": "myupgrade",
                "command": "pytest tests -m 'not integration'",
            },
            {
                "name": "myenv_lower",
                "lower": "mylower",
                "command": "pytest tests -m 'not integration'",
            },
        ]
    }

    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(toml)


def test_parse_toml_default(tmpdir):
    """Test parsing a config with no install requirements and defaults."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "myconfig.toml")
    with open(conf_loc, "w") as outfile:
        outfile.write(TOML_NOREQS_DEFAULTS)

    toml = parse_toml(filename=conf_loc)

    assert toml == {
        "envs": [
            {
                "name": "myenv",
                "upgrade": "myupgrade",
                "extras": "tests",
                "command": "pytest tests",
            },
            {
                "name": "myenv_lower",
                "lower": "mylower",
                "extras": "tests",
                "command": "pytest tests",
            },
        ]
    }

    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(toml)


def test_parse_toml_reqs(tmpdir):
    """Test parsing a TOML style config."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "pyproject.toml")
    with open(conf_loc, "w") as outfile:
        outfile.write(TOML_REQS)

    toml = parse_toml(filename=conf_loc)

    assert toml == {
        "envs": [
            {"name": "myupgrade", "upgrade": "myupgrade"},
            {"name": "all-requirements", "upgrade": "myupgrade"},
        ]
    }

    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(toml)


def test_parse_toml_reqs_default(tmpdir):
    """Test parsing a TOML style config with default arguments."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "pyproject.toml")
    with open(conf_loc, "w") as outfile:
        outfile.write(TOML_REQS_DEFAULTS)

    toml = parse_toml(filename=conf_loc)

    assert toml == {
        "envs": [
            {
                "name": "myupgrade",
                "upgrade": "myupgrade",
                "extras": "tests",
                "command": "pytest tests -m 'not integration'",
            },
            {
                "name": "all-requirements",
                "upgrade": "myupgrade",
                "extras": "tests",
                "command": "pytest tests -m 'not integration'",
            },
        ]
    }

    validator = EdgetestValidator(schema=BASE_SCHEMA)

    assert validator.validate(toml)


def test_parse_custom_toml(tmpdir):
    """Test parsing a custom configuration."""
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "pyproject.toml")
    with open(conf_loc, "w") as outfile:
        outfile.write(TOML_CUSTOM)

    toml = parse_toml(filename=conf_loc)

    assert toml == {
        "custom": {"mycustom": "mykey"},
        "envs": [
            {
                "name": "myenv",
                "upgrade": "myupgrade",
                "extras": "tests",
                "command": "pytest tests -m 'not integration'",
            },
            {
                "name": "myenv_lower",
                "lower": "mylower",
                "extras": "tests",
                "command": "pytest tests -m 'not integration'",
            },
        ],
    }

    schema = Schema()
    schema.add_globaloption(
        "custom", {"type": "dict", "schema": {"mycustom": {"type": "string"}}}
    )

    validator = EdgetestValidator(schema=schema.schema)

    assert validator.validate(toml)


def test_convert_toml_array_to_string():
    test_array = array("['a','b','c', 'd']")
    test_string = string("abcd")

    assert _convert_toml_array_to_string(test_array) == "a\nb\nc\nd"
    assert _convert_toml_array_to_string(test_string) == "abcd"


def test_upgrade_setup_cfg(tmpdir):
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "setup.cfg")
    with open(conf_loc, "w") as outfile:
        outfile.write(CFG_REQS_UPGRADE)

    upgrade_cfg = upgrade_setup_cfg(
        upgraded_packages=[
            {"name": "pandas", "version": "2.0.0"},
            {"name": "numpy", "version": "3.0.0"},
            {"name": "pytest", "version": "4.0.0"},
        ],
        filename=conf_loc,
    )

    assert (
        upgrade_cfg["options"]["install_requires"]
        == "\npandas<=2.0.0,>=1.0.0\nnumpy<=3.0.0,>=1.0.0"
    )
    assert upgrade_cfg["options.extras_require"]["tests"] == "\npytest<=4.0.0,>=1.0.0"


def test_upgrade_pyproject_toml(tmpdir):
    location = tmpdir.mkdir("mylocation")
    conf_loc = Path(str(location), "pyproject.toml")
    with open(conf_loc, "w") as outfile:
        outfile.write(TOML_REQS_UPGRADE)

    assert upgrade_pyproject_toml(
        upgraded_packages=[
            {"name": "pandas", "version": "2.0.0"},
            {"name": "numpy", "version": "3.0.0"},
            {"name": "pytest", "version": "4.0.0"},
        ],
        filename=conf_loc,
    ) == {
        "project": {
            "dependencies": ["pandas<=2.0.0,>=1.0.0", "numpy<=3.0.0,>=1.0.0"],
            "optional-dependencies": {"tests": ["pytest<=4.0.0,>=1.0.0"]},
        }
    }


def test_isin_case_dashhyphen_ins():
    vals = ["pandas", "python-dateutil"]

    assert _isin_case_dashhyphen_ins("pandas", vals)
    assert _isin_case_dashhyphen_ins("Pandas", vals)
    assert not _isin_case_dashhyphen_ins("Panda$", vals)
    assert _isin_case_dashhyphen_ins("python-dateutil", vals)
    assert _isin_case_dashhyphen_ins("Python_Dateutil", vals)
    assert not _isin_case_dashhyphen_ins("Python_Dateut1l", vals)
    assert not _isin_case_dashhyphen_ins("pandaspython-dateutil", vals)



"""Test the CLI."""

import platform
from pathlib import Path
from unittest.mock import PropertyMock, call, patch

from click.testing import CliRunner

from edgetest.interface import cli

CURR_DIR = Path(__file__).resolve().parent

REQS = """
myupgrade
"""

SETUP_TOML = """
[edgetest.envs.myenv]
upgrade = ["myupgrade"]
command = "pytest tests -m 'not integration'"
"""

SETUP_TOML_LOWER = """
[project]
dependencies = ["myupgrade<=0.1.5", "mylower<=0.1.0,>=0.0.1"]

[edgetest.envs.myenv_lower]
lower = ["mylower"]
command = "pytest tests -m 'not integration'"
"""

SETUP_TOML_REQS = """
[project]
dependencies = ["myupgrade<=0.1.5"]
"""

SETUP_TOML_REQS_UPGRADE = """
[project]
dependencies = ["myupgrade<=0.2.0"]
"""

SETUP_TOML_EXTRAS = """
[project.optional-dependencies]
myextra = ["myupgrade<=0.1.5"]

[edgetest.envs.myenv]
upgrade = ["myupgrade"]
extras = ["myextra"]
command = "pytest tests -m 'not integration'"
"""


SETUP_TOML_EXTRAS_UPGRADE = """
[project.optional-dependencies]
myextra = ["myupgrade<=0.2.0"]

[edgetest.envs.myenv]
upgrade = ["myupgrade"]
extras = ["myextra"]
command = "pytest tests -m 'not integration'"
"""

PIP_LIST = """
[{"name": "myupgrade", "version": "0.2.0"}]
"""

TABLE_OUTPUT = """

=============  ==================  ===============  ===================  ==================  =================
Environment    Setup successful    Passing tests    Upgraded packages    Lowered packages    Package version
=============  ==================  ===============  ===================  ==================  =================
myenv          True                True             myupgrade                                0.2.0
=============  ==================  ===============  ===================  ==================  =================
"""

TABLE_OUTPUT_LOWER = """

=============  ==================  ===============  ===================  ==================  =================
Environment    Setup successful    Passing tests    Upgraded packages    Lowered packages    Package version
=============  ==================  ===============  ===================  ==================  =================
myenv_lower    True                True                                  mylower             0.0.1
=============  ==================  ===============  ===================  ==================  =================
"""

TABLE_OUTPUT_NOTEST = """

=============  ==================  ===============  ===================  ==================  =================
Environment    Setup successful    Passing tests    Upgraded packages    Lowered packages    Package version
=============  ==================  ===============  ===================  ==================  =================
myenv          True                False            myupgrade                                0.2.0
=============  ==================  ===============  ===================  ==================  =================
"""

TABLE_OUTPUT_NOTEST_LOWER = """

=============  ==================  ===============  ===================  ==================  =================
Environment    Setup successful    Passing tests    Upgraded packages    Lowered packages    Package version
=============  ==================  ===============  ===================  ==================  =================
myenv_lower    True                False                                 mylower             0.0.1
=============  ==================  ===============  ===================  ==================  =================
"""

TABLE_OUTPUT_REQS = """

================  ==================  ===============  ===================  ==================  =================
Environment       Setup successful    Passing tests    Upgraded packages    Lowered packages    Package version
================  ==================  ===============  ===================  ==================  =================
myupgrade         True                True             myupgrade                                0.2.0
all-requirements  True                True             myupgrade                                0.2.0
================  ==================  ===============  ===================  ==================  =================
"""


@patch("edgetest.lib.EnvBuilder", autospec=True)
@patch("edgetest.core.Popen", autospec=True)
@patch("edgetest.utils.Popen", autospec=True)
def test_cli_basic(mock_popen, mock_cpopen, mock_builder):
    """Test creating a basic environment."""
    mock_popen.return_value.communicate.return_value = (PIP_LIST, "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=0)
    mock_cpopen.return_value.communicate.return_value = ("output", "error")
    type(mock_cpopen.return_value).returncode = PropertyMock(return_value=0)

    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("pyproject.toml", "w") as outfile:
            outfile.write(SETUP_TOML)

        result = runner.invoke(cli, ["--config=pyproject.toml"])

    assert result.exit_code == 0

    env_loc = Path(loc) / ".edgetest" / "myenv"
    if platform.system() == "Windows":
        py_loc = env_loc / "Scripts" / "python"
    else:
        py_loc = env_loc / "bin" / "python"

    mock_builder.return_value.create.assert_called_with(env_loc)
    assert mock_popen.call_args_list == [
        call(
            (f"{str(py_loc)}", "-m", "pip", "install", "."),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (
                f"{str(py_loc)}",
                "-m",
                "pip",
                "install",
                "myupgrade",
                "--upgrade",
            ),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (f"{str(py_loc)}", "-m", "pip", "list", "--format", "json"),
            stdout=-1,
            universal_newlines=True,
        ),
    ]
    assert mock_cpopen.call_args_list == [
        call(
            (
                f"{str(py_loc)}",
                "-m",
                "pytest",
                "tests",
                "-m",
                "not integration",
            ),
            universal_newlines=True,
        )
    ]

    assert result.output == TABLE_OUTPUT


@patch("edgetest.lib.EnvBuilder", autospec=True)
@patch("edgetest.core.Popen", autospec=True)
@patch("edgetest.utils.Popen", autospec=True)
def test_cli_basic_lower(mock_popen, mock_cpopen, mock_builder):
    """Test creating a basic environment."""
    mock_popen.return_value.communicate.return_value = (PIP_LIST, "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=0)
    mock_cpopen.return_value.communicate.return_value = ("output", "error")
    type(mock_cpopen.return_value).returncode = PropertyMock(return_value=0)

    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("pyproject.toml", "w") as outfile:
            outfile.write(SETUP_TOML_LOWER)

        result = runner.invoke(cli, ["--config=pyproject.toml"])

    assert result.exit_code == 0

    env_loc = Path(loc) / ".edgetest" / "myenv_lower"
    if platform.system() == "Windows":
        py_loc = env_loc / "Scripts" / "python"
    else:
        py_loc = env_loc / "bin" / "python"

    mock_builder.return_value.create.assert_called_with(env_loc)
    assert mock_popen.call_args_list == [
        call(
            (f"{str(py_loc)}", "-m", "pip", "install", "."),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (
                f"{str(py_loc)}",
                "-m",
                "pip",
                "install",
                "mylower==0.0.1",
            ),
            stdout=-1,
            universal_newlines=True,
        ),
    ]
    assert mock_cpopen.call_args_list == [
        call(
            (
                f"{str(py_loc)}",
                "-m",
                "pytest",
                "tests",
                "-m",
                "not integration",
            ),
            universal_newlines=True,
        )
    ]

    assert result.output == TABLE_OUTPUT_LOWER


@patch("edgetest.lib.EnvBuilder", autospec=True)
@patch("edgetest.core.Popen", autospec=True)
@patch("edgetest.utils.Popen", autospec=True)
def test_cli_reqs(mock_popen, mock_cpopen, mock_builder):
    """Test running tests based on the requirements file."""
    mock_popen.return_value.communicate.return_value = (PIP_LIST, "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=0)
    mock_cpopen.return_value.communicate.return_value = ("output", "error")
    type(mock_cpopen.return_value).returncode = PropertyMock(return_value=0)

    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("requirements.txt", "w") as outfile:
            outfile.write(REQS)

        result = runner.invoke(cli)

    if platform.system() == "Windows":
        py_myupgrade_loc = Path(loc) / ".edgetest" / "myupgrade" / "Scripts" / "python"
        py_allreq_loc = (
            Path(loc) / ".edgetest" / "all-requirements" / "Scripts" / "python"
        )
    else:
        py_myupgrade_loc = Path(loc) / ".edgetest" / "myupgrade" / "bin" / "python"
        py_allreq_loc = Path(loc) / ".edgetest" / "all-requirements" / "bin" / "python"

    assert result.exit_code == 0

    assert mock_builder.return_value.create.call_args_list == [
        call(env_dir=Path(loc) / ".edgetest" / "myupgrade"),
        call(env_dir=Path(loc) / ".edgetest" / "all-requirements"),
    ]

    assert mock_popen.call_args_list == [
        call(
            (f"{str(py_myupgrade_loc)}", "-m", "pip", "install", "."),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (
                f"{str(py_myupgrade_loc)}",
                "-m",
                "pip",
                "install",
                "myupgrade",
                "--upgrade",
            ),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (
                f"{str(py_allreq_loc)}",
                "-m",
                "pip",
                "install",
                ".",
            ),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (
                f"{str(py_allreq_loc)}",
                "-m",
                "pip",
                "install",
                "myupgrade",
                "--upgrade",
            ),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (
                f"{str(py_myupgrade_loc)}",
                "-m",
                "pip",
                "list",
                "--format",
                "json",
            ),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (
                f"{str(py_allreq_loc)}",
                "-m",
                "pip",
                "list",
                "--format",
                "json",
            ),
            stdout=-1,
            universal_newlines=True,
        ),
    ]
    assert mock_cpopen.call_args_list == [
        call(
            (f"{str(py_myupgrade_loc)}", "-m", "pytest"),
            universal_newlines=True,
        ),
        call(
            (f"{str(py_allreq_loc)}", "-m", "pytest"),
            universal_newlines=True,
        ),
    ]

    assert result.output == TABLE_OUTPUT_REQS


@patch("edgetest.lib.EnvBuilder", autospec=True)
@patch("edgetest.core.Popen", autospec=True)
@patch("edgetest.utils.Popen", autospec=True)
def test_cli_setup_reqs_update(mock_popen, mock_cpopen, mock_builder):
    """Test running tests and updating requirements in a ``pyproject.toml`` file."""
    mock_popen.return_value.communicate.return_value = (PIP_LIST, "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=0)
    mock_cpopen.return_value.communicate.return_value = ("output", "error")
    type(mock_cpopen.return_value).returncode = PropertyMock(return_value=0)

    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("pyproject.toml", "w") as outfile:
            outfile.write(SETUP_TOML_REQS)

        result = runner.invoke(cli, ["--config=pyproject.toml", "--export"])

        with open("pyproject.toml") as infile:
            out = infile.read()

    assert result.exit_code == 0

    assert out == SETUP_TOML_REQS_UPGRADE


@patch("edgetest.lib.EnvBuilder", autospec=True)
@patch("edgetest.core.Popen", autospec=True)
@patch("edgetest.utils.Popen", autospec=True)
def test_cli_setup_extras_update(mock_popen, mock_cpopen, mock_builder):
    """Test running tests and updating extra installation requirements in a ``pyproject.toml`` file."""
    mock_popen.return_value.communicate.return_value = (PIP_LIST, "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=0)
    mock_cpopen.return_value.communicate.return_value = ("output", "error")
    type(mock_cpopen.return_value).returncode = PropertyMock(return_value=0)

    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("pyproject.toml", "w") as outfile:
            outfile.write(SETUP_TOML_EXTRAS)

        result = runner.invoke(cli, ["--config=pyproject.toml", "--export"])

        print(result.output)

        with open("pyproject.toml") as infile:
            out = infile.read()

    assert result.exit_code == 0

    assert out == SETUP_TOML_EXTRAS_UPGRADE


@patch("edgetest.core.Popen", autospec=True)
@patch("edgetest.utils.Popen", autospec=True)
def test_cli_nosetup(mock_popen, mock_cpopen):
    """Test creating a basic environment."""
    mock_popen.return_value.communicate.return_value = (PIP_LIST, "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=0)
    mock_cpopen.return_value.communicate.return_value = ("output", "error")
    type(mock_cpopen.return_value).returncode = PropertyMock(return_value=0)

    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("pyproject.toml", "w") as outfile:
            outfile.write(SETUP_TOML)

        result = runner.invoke(cli, ["--config=pyproject.toml", "--nosetup"])

    assert result.exit_code == 0

    env_loc = str(Path(loc) / ".edgetest" / "myenv")
    if platform.system() == "Windows":
        py_loc = Path(env_loc) / "Scripts" / "python"
    else:
        py_loc = Path(env_loc) / "bin" / "python"

    assert mock_popen.call_args_list == [
        call(
            (f"{py_loc}", "-m", "pip", "list", "--format", "json"),
            stdout=-1,
            universal_newlines=True,
        ),
    ]
    assert mock_cpopen.call_args_list == [
        call(
            (f"{py_loc}", "-m", "pytest", "tests", "-m", "not integration"),
            universal_newlines=True,
        )
    ]

    assert (
        result.output == f"""Using existing environment for myenv...\n{TABLE_OUTPUT}"""
    )


@patch("edgetest.core.Popen", autospec=True)
@patch("edgetest.utils.Popen", autospec=True)
def test_cli_nosetup_lower(mock_popen, mock_cpopen):
    """Test creating a basic environment."""
    mock_popen.return_value.communicate.return_value = (PIP_LIST, "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=0)
    mock_cpopen.return_value.communicate.return_value = ("output", "error")
    type(mock_cpopen.return_value).returncode = PropertyMock(return_value=0)

    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("pyproject.toml", "w") as outfile:
            outfile.write(SETUP_TOML_LOWER)

        result = runner.invoke(cli, ["--config=pyproject.toml", "--nosetup"])

    assert result.exit_code == 0

    env_loc = str(Path(loc) / ".edgetest" / "myenv_lower")
    if platform.system() == "Windows":
        py_loc = Path(env_loc) / "Scripts" / "python"
    else:
        py_loc = Path(env_loc) / "bin" / "python"

    assert mock_cpopen.call_args_list == [
        call(
            (f"{py_loc}", "-m", "pytest", "tests", "-m", "not integration"),
            universal_newlines=True,
        )
    ]

    assert (
        result.output
        == f"""Using existing environment for myenv_lower...\n{TABLE_OUTPUT_LOWER}"""
    )


@patch("edgetest.lib.EnvBuilder", autospec=True)
@patch("edgetest.utils.Popen", autospec=True)
def test_cli_notest(mock_popen, mock_builder):
    """Test creating a basic environment."""
    mock_popen.return_value.communicate.return_value = (PIP_LIST, "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=0)

    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("pyproject.toml", "w") as outfile:
            outfile.write(SETUP_TOML)

        result = runner.invoke(cli, ["--config=pyproject.toml", "--notest"])

    assert result.exit_code == 0

    env_loc = Path(loc) / ".edgetest" / "myenv"
    if platform.system() == "Windows":
        py_loc = env_loc / "Scripts" / "python"
    else:
        py_loc = env_loc / "bin" / "python"

    mock_builder.return_value.create.assert_called_with(env_loc)
    assert mock_popen.call_args_list == [
        call(
            (f"{str(py_loc)}", "-m", "pip", "install", "."),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (
                f"{str(py_loc)}",
                "-m",
                "pip",
                "install",
                "myupgrade",
                "--upgrade",
            ),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (f"{str(py_loc)}", "-m", "pip", "list", "--format", "json"),
            stdout=-1,
            universal_newlines=True,
        ),
    ]

    assert result.output == f"""Skipping tests for myenv\n{TABLE_OUTPUT_NOTEST}"""


@patch("edgetest.lib.EnvBuilder", autospec=True)
@patch("edgetest.utils.Popen", autospec=True)
def test_cli_notest_lower(mock_popen, mock_builder):
    """Test creating a basic environment."""
    mock_popen.return_value.communicate.return_value = (PIP_LIST, "error")
    type(mock_popen.return_value).returncode = PropertyMock(return_value=0)

    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("pyproject.toml", "w") as outfile:
            outfile.write(SETUP_TOML_LOWER)

        result = runner.invoke(cli, ["--config=pyproject.toml", "--notest"])

    assert result.exit_code == 0

    env_loc = Path(loc) / ".edgetest" / "myenv_lower"
    if platform.system() == "Windows":
        py_loc = env_loc / "Scripts" / "python"
    else:
        py_loc = env_loc / "bin" / "python"

    mock_builder.return_value.create.assert_called_with(env_loc)
    assert mock_popen.call_args_list == [
        call(
            (f"{str(py_loc)}", "-m", "pip", "install", "."),
            stdout=-1,
            universal_newlines=True,
        ),
        call(
            (
                f"{str(py_loc)}",
                "-m",
                "pip",
                "install",
                "mylower==0.0.1",
            ),
            stdout=-1,
            universal_newlines=True,
        ),
    ]

    assert (
        result.output
        == f"""Skipping tests for myenv_lower\n{TABLE_OUTPUT_NOTEST_LOWER}"""
    )



"""Create a fake package and test."""
import configparser
import sys
from pathlib import Path

import pytest
from click.testing import CliRunner
from tomlkit import load

from edgetest.interface import cli

SETUP_TOML = """
[project]
name = "toy_package"
version = "0.1.0"
description = "Fake description"
requires-python = ">=3.7.0"
dependencies = ["pandas<=1.2.0"]

[project.optional-dependencies]
tests = ["pytest"]

[edgetest]
extras = ["tests"]
"""

SETUP_TOML_LOWER = """
[project]
name = "toy_package"
version = "0.1.0"
description = "Fake description"
requires-python = ">=3.7.0"
dependencies = ["pandas<=1.2.0,>=1.0.0"]

[project.optional-dependencies]
tests = ["pytest"]

[edgetest]
extras = ["tests"]

[edgetest.envs.lower_env]
lower = ["pandas"]
"""

SETUP_TOML_DASK = """
[project]
name = "toy_package"
version = "0.1.0"
description = "Fake description"
requires-python = ">=3.7.0"
dependencies = ["Scikit_Learn>=1.0,<=1.2.0", "Dask[dataframe]<=2022.1.0,>=2021.6.1"]

[project.optional-dependencies]
tests = ["pytest"]

[edgetest.envs.core]
extras = ["tests"]
upgrade = ["scikit-learn", "dask[dataframe]"]

[edgetest.envs.lower_env]
extras = ["tests"]
lower = ["scikit-learn", "dask[dataframe]"]
"""


SETUP_PY = """
from setuptools import setup

setup()
"""

MODULE_CODE = """
def main() -> str:
    print("Hello world")
"""

TEST_CODE = """
from toy_package import main

def test_func():
    main()
"""

PY_VER = f"python{sys.version_info.major}.{sys.version_info.minor}"


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
@pytest.mark.integration
def test_toy_package():
    """Test using edgetest with a toy package."""
    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("pyproject.toml", "w") as outfile:
            outfile.write(SETUP_TOML)
        with open("setup.py", "w") as outfile:
            outfile.write(SETUP_PY)
        # Make a directory for the module
        Path(loc, "toy_package").mkdir()
        with open(Path(loc, "toy_package", "__init__.py"), "w") as outfile:
            outfile.write(MODULE_CODE)
        # Make a directory for the tests
        Path(loc, "tests").mkdir()
        with open(Path(loc, "tests", "test_main.py"), "w") as outfile:
            outfile.write(TEST_CODE)

        # Run the CLI
        result = runner.invoke(cli, ["--config=pyproject.toml"])

        assert result.exit_code == 0
        assert Path(loc, ".edgetest").is_dir()
        assert Path(loc, ".edgetest", "pandas").is_dir()
        assert "pandas" in result.stdout

        if not sys.platform == "win32":
            assert Path(
                loc, ".edgetest", "pandas", "lib", PY_VER, "site-packages", "pandas"
            ).is_dir()


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
@pytest.mark.integration
def test_toy_package_lower():
    """Test using edgetest with a toy package."""
    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("pyproject.toml", "w") as outfile:
            outfile.write(SETUP_TOML_LOWER)
        with open("setup.py", "w") as outfile:
            outfile.write(SETUP_PY)
        # Make a directory for the module
        Path(loc, "toy_package").mkdir()
        with open(Path(loc, "toy_package", "__init__.py"), "w") as outfile:
            outfile.write(MODULE_CODE)
        # Make a directory for the tests
        Path(loc, "tests").mkdir()
        with open(Path(loc, "tests", "test_main.py"), "w") as outfile:
            outfile.write(TEST_CODE)

        # Run the CLI
        result = runner.invoke(cli, ["--config=pyproject.toml"])

        assert result.exit_code == 0
        assert Path(loc, ".edgetest").is_dir()
        assert Path(loc, ".edgetest", "lower_env").is_dir()
        assert "pandas" in result.stdout

        if not sys.platform == "win32":
            assert Path(
                loc, ".edgetest", "lower_env", "lib", PY_VER, "site-packages", "pandas"
            ).is_dir()


@pytest.mark.integration
def test_toy_package_dask():
    """Test using edgetest with a toy package."""
    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("pyproject.toml", "w") as outfile:
            outfile.write(SETUP_TOML_DASK)
        with open("setup.py", "w") as outfile:
            outfile.write(SETUP_PY)
        # Make a directory for the module
        Path(loc, "toy_package").mkdir()
        with open(Path(loc, "toy_package", "__init__.py"), "w") as outfile:
            outfile.write(MODULE_CODE)
        # Make a directory for the tests
        Path(loc, "tests").mkdir()
        with open(Path(loc, "tests", "test_main.py"), "w") as outfile:
            outfile.write(TEST_CODE)

        # Run the CLI
        result = runner.invoke(cli, ["--config=pyproject.toml", "--export"])

        assert result.exit_code == 0
        assert Path(loc, ".edgetest").is_dir()

        for envname in ("core", "lower_env"):
            assert Path(loc, ".edgetest", envname).is_dir()

            if not sys.platform == "win32":
                assert Path(
                    loc, ".edgetest", envname, "lib", PY_VER, "site-packages", "dask"
                ).is_dir()
                assert Path(
                    loc, ".edgetest", envname, "lib", PY_VER, "site-packages", "pandas"
                ).is_dir()
                assert Path(
                    loc, ".edgetest", envname, "lib", PY_VER, "site-packages", "sklearn"
                ).is_dir()

        config = load(open("pyproject.toml"))

        assert "Scikit_Learn" in config["project"]["dependencies"][0]
        assert "Dask" in config["project"]["dependencies"][1]
        assert config["edgetest"]["envs"]["core"]["extras"] == ["tests"]
        assert config["edgetest"]["envs"]["core"]["upgrade"] == [
            "scikit-learn",
            "dask[dataframe]",
        ]
        assert config["edgetest"]["envs"]["lower_env"]["extras"] == ["tests"]
        assert config["edgetest"]["envs"]["lower_env"]["lower"] == [
            "scikit-learn",
            "dask[dataframe]",
        ]
        assert "dask" in result.stdout
        assert "scikit-learn" in result.stdout



"""Define the Cerberus schema for the testing configuration."""

from typing import Dict, List

from cerberus import Validator

BASE_SCHEMA = {
    "envs": {
        "type": "list",
        "required": True,
        "schema": {
            "type": "dict",
            "schema": {
                "name": {"type": "string", "coerce": "strip", "required": True},
                "upgrade": {
                    "type": "list",
                    "schema": {
                        "type": "string",
                    },
                    "coerce": "listify",
                    "required": True,
                    "excludes": "lower",
                },
                "lower": {
                    "type": "list",
                    "schema": {"type": "string"},
                    "coerce": "listify",
                    "required": True,
                    "excludes": "upgrade",
                },
                "extras": {
                    "type": "list",
                    "schema": {"type": "string"},
                    "coerce": "listify",
                    "default": None,
                    "nullable": True,
                },
                "deps": {
                    "type": "list",
                    "schema": {"type": "string"},
                    "coerce": "listify",
                    "default": None,
                    "nullable": True,
                },
                "command": {"type": "string", "coerce": "strip", "default": "pytest"},
                "package_dir": {"type": "string", "coerce": "strip", "default": "."},
            },
        },
    }
}


class Schema:
    """An editable schema."""

    schema = BASE_SCHEMA

    def add_envoption(self, option: str, schema: Dict):
        """Add an environment-level option.

        Parameters
        ----------
        option : str
            The name of the option. This will be the key in the key-value pair.
        schema : dict
            The schema for the option.

        Examples
        --------
        >>> Schema().add_envoption("command", {"type": "string", "default": "pytest"})
        """
        self.schema["envs"]["schema"]["schema"][option] = schema  # type: ignore

    def add_globaloption(self, option: str, schema: Dict):
        """Add a global option.

        Parameters
        ----------
        option : str
            The name of the option. This will be the key in the key-value pair.
        schema : dict
            The schema for the option.

        Examples
        --------
        >>> Schema().add_globaloption("print_message", {"type": "string"})
        """
        self.schema[option] = schema


class EdgetestValidator(Validator):
    """Custom validator for coercing lists from ``.ini`` style files."""

    def _normalize_coerce_listify(self, value: str) -> List:
        """Coerce a value into a list.

        Parameters
        ----------
        value : str
            The original value for the field.

        Returns
        -------
        list
            The newline-separated list.
        """
        if isinstance(value, str):
            return value.strip().splitlines()
        else:
            return value

    def _normalize_coerce_strip(self, value: str) -> str:
        """Remove leading and trailing spaces.

        Parameters
        ----------
        value : str
            The original value for the field.

        Returns
        -------
        str
            The stripped string.
        """
        return value.strip()



"""Core module."""

import json
import shlex
from pathlib import Path
from subprocess import Popen
from typing import Dict, List, Optional

from pluggy._hooks import _HookRelay

from .logger import get_logger
from .utils import _isin_case_dashhyphen_ins, _run_command, pushd

LOG = get_logger(__name__)


class TestPackage:
    """Run test commands with bleeding edge dependencies.

    Parameters
    ----------
    hook : _HookRelay
        The hook object from ``pluggy``.
    envname : str
        The name of the conda environment to create/use.
    upgrade : list
        The list of packages to upgrade
    lower : list
        The list of packages to install lower bounds alongside the lower bound value.
        E.g. ``["pandas==1.5.2"]``.
    package_dir : str, optional (default None)
        The location of the local package to install and test.

    Attributes
    ----------
    _basedir : str
        The base directory location for each environment
    status : bool
        A boolean status indicator for whether or not the tests passed. Only populated
        after ``run_tests`` has been executed.
    """

    def __init__(
        self,
        hook: _HookRelay,
        envname: str,
        upgrade: Optional[List[str]] = None,
        lower: Optional[List[str]] = None,
        package_dir: Optional[str] = None,
    ):
        """Init method."""
        self.hook = hook
        self._basedir = Path(Path.cwd(), ".edgetest")
        self._basedir.mkdir(exist_ok=True)
        self.envname = envname

        if (upgrade is None and lower is None) or (
            upgrade is not None and lower is not None
        ):
            raise ValueError("Exactly one of upgrade and lower must be provided.")
        self.upgrade = upgrade
        self.lower = lower
        self.package_dir = package_dir or "."

        self.setup_status: bool = False
        self.status: bool = False

    @property
    def python_path(self) -> str:
        """Get the path to the python executable.

        Returns
        -------
        str
            The path to the python executable.
        """
        return self.hook.path_to_python(basedir=self._basedir, envname=self.envname)  # type: ignore

    def setup(
        self,
        extras: Optional[List[str]] = None,
        deps: Optional[List[str]] = None,
        skip: bool = False,
        **options,
    ) -> None:
        """Set up the testing environment.

        Parameters
        ----------
        extras : list, optional (default None)
            The list of extra installations to include.
        deps : list, optional (default None)
            A list of additional dependencies to install via ``pip``
        skip : bool, optional (default False)
            Whether to skip setup as a pre-made environment has already been
            created.
        **options
            Additional options for ``self.hook.create_environment``.

        Returns
        -------
        None

        Raises
        ------
        RuntimeError
            This error will be raised if any part of the set up process fails.
        """
        if skip:
            self.setup_status = True
            return
        # Create the conda environment
        try:
            LOG.info(f"Creating the following environment: {self.envname}...")
            self.hook.create_environment(
                basedir=self._basedir, envname=self.envname, conf=options
            )
            LOG.info(f"Successfully created {self.envname}")
        except RuntimeError:
            LOG.exception(
                "Could not create the following environment: %s", self.envname
            )
            self.setup_status = False
            return

        # Install the local package
        with pushd(self.package_dir):
            pkg = "."
            if extras:
                pkg += f"[{', '.join(extras)}]"
            if deps:
                LOG.info(
                    "Installing specified additional dependencies into %s: %s",
                    self.envname,
                    ", ".join(deps),
                )
                split = [shlex.split(dep) for dep in deps]
                try:
                    _run_command(
                        self.python_path,
                        "-m",
                        "pip",
                        "install",
                        *[itm for lst in split for itm in lst],
                    )
                except RuntimeError:
                    LOG.exception(
                        "Unable to install specified additional dependencies in %s",
                        self.envname,
                    )
                    self.setup_status = False
                    return
                LOG.info(
                    f"Successfully installed specified additional dependencies into {self.envname}"
                )

            LOG.info(f"Installing the local package into {self.envname}...")
            try:
                _run_command(self.python_path, "-m", "pip", "install", pkg)
                LOG.info(
                    f"Successfully installed the local package into {self.envname}..."
                )
            except RuntimeError:
                LOG.exception(
                    "Unable to install the local package into %s", self.envname
                )
                self.setup_status = False
                return

        if self.upgrade:
            # Upgrade package(s)
            LOG.info(
                f"Upgrading the following packages in {self.envname}: {', '.join(self.upgrade)}"
            )
            try:
                self.hook.run_update(
                    basedir=self._basedir,
                    envname=self.envname,
                    upgrade=self.upgrade,
                    conf=options,
                )
                self.setup_status = True
            except RuntimeError:
                self.setup_status = False
                LOG.exception("Unable to upgrade packages in %s", self.envname)
                return
            LOG.info("Successfully upgraded packages in %s", self.envname)
        else:
            # Install lower bounds of package(s)
            LOG.info(
                "Installing lower bounds of packages in %s: %s",
                {self.envname},
                ", ".join(self.lower),  # type:ignore
            )
            try:
                self.hook.run_install_lower(
                    basedir=self._basedir,
                    envname=self.envname,
                    lower=self.lower,
                    conf=options,
                )
                self.setup_status = True
            except RuntimeError:
                self.setup_status = False
                LOG.exception(
                    "Unable to install lower bounds of packages in %s", self.envname
                )
                return
            LOG.info(
                f"Successfully installed lower bounds of packages in {self.envname}"
            )

    def upgraded_packages(self) -> List[Dict[str, str]]:
        """Get the list of upgraded packages for the test environment.

        Parameters
        ----------
        None

        Returns
        -------
        List
            The output of ``pip list --format json``, filtered to the packages upgraded
            for this environment.
        """
        if self.upgrade is None:
            return []
        # Get the version for the upgraded package(s)
        out, _ = _run_command(self.python_path, "-m", "pip", "list", "--format", "json")
        outjson = json.loads(out)

        upgrade_wo_extras = [pkg.split("[")[0] for pkg in self.upgrade]
        return [
            pkg
            for pkg in outjson
            if _isin_case_dashhyphen_ins(pkg.get("name", ""), upgrade_wo_extras)
        ]

    def lowered_packages(self) -> List[Dict[str, str]]:
        """Get list of lowered packages for the test environment.

        Returns
        -------
        List[Dict[str, str]]
            List of lowered packages and their versions
        """
        if self.lower is None:
            return []
        packages_split = (pkg_str.split("==") for pkg_str in self.lower)
        return [
            {"name": pkg_info[0], "version": pkg_info[1]} for pkg_info in packages_split
        ]

    def run_tests(self, command: str) -> int:
        """Run the tests in the package directory.

        Parameters
        ----------
        command : str
            The test command

        Returns
        -------
        int
            The exit code
        """
        if not self.setup_status:
            raise RuntimeError("Environment setup failed. Cannot run tests.")
        with pushd(self.package_dir):
            popen = Popen(
                (self.python_path, "-m", *shlex.split(command)), universal_newlines=True
            )
            popen.communicate()

        self.status = bool(popen.returncode == 0)

        return popen.returncode



from pathlib import Path
from unittest.mock import patch

import pytest

from edgetest.lib import (
    create_environment,
    path_to_python,
    run_install_lower,
    run_update,
)


@patch("edgetest.lib.platform", autospec=True)
def test_path_to_python(mock_platform):
    mock_platform.system.return_value = "Windows"
    assert path_to_python("test", "test") == str(
        Path("test") / "test" / "Scripts" / "python"
    )

    mock_platform.system.return_value = "Unix"
    assert path_to_python("test", "test") == str(
        Path("test") / "test" / "bin" / "python"
    )

    mock_platform.system.side_effect = RuntimeError()
    with pytest.raises(RuntimeError):
        path_to_python("test", "test")


@patch("edgetest.lib.EnvBuilder", autospec=True)
def test_create_environment(mock_env_builder):
    create_environment("test", "test", {})
    mock_env_builder.assert_called_with(with_pip=True)
    mock_env_builder().create.assert_called_with(env_dir=Path("test", "test"))

    mock_env_builder().create.side_effect = RuntimeError()
    with pytest.raises(RuntimeError):
        create_environment("test", "test", {})


@patch("edgetest.lib._run_command", autospec=True)
def test_run_update(mock_run):
    python_path = path_to_python("test", "test")
    run_update("test", "test", ["1", "2"], {"test": "test"})
    mock_run.assert_called_with(
        python_path, "-m", "pip", "install", "1", "2", "--upgrade"
    )

    mock_run.side_effect = RuntimeError()
    with pytest.raises(RuntimeError):
        run_update("test", "test", ["1", "2"], {"test": "test"})


@patch("edgetest.lib._run_command", autospec=True)
def test_run_install_lower(mock_run):
    python_path = path_to_python("test", "test")
    run_install_lower("test", "test", ["package1==1", "package2==2"], {"test": "test"})
    mock_run.assert_called_with(
        python_path,
        "-m",
        "pip",
        "install",
        "package1==1",
        "package2==2",
    )

    mock_run.side_effect = RuntimeError()
    with pytest.raises(RuntimeError):
        run_install_lower(
            "test", "test", ["package1==1", "package2==2"], {"test": "test"}
        )



"""Package initialization."""


__version__ = "2024.2.0"

__title__ = "edgetest"
__description__ = "Bleeding edge dependency testing"
__url__ = "https://github.com/capitalone/edgetest"
__uri__ = __url__
__doc__ = __description__ + " <" + __uri__ + ">"

__author__ = "Akshay Gupta"
__email__ = "akshay.gupta2@capitalone.com"



"""Generate rST reports."""

from typing import Any, List

from tabulate import tabulate

from .core import TestPackage

VALID_OUTPUTS = ["rst", "github"]


def gen_report(testers: List[TestPackage], output_type: str = "rst") -> Any:
    """Generate a rST report.

    Parameters
    ----------
    testers : list
        A list of ``TestPackage`` objects.

    output_type : str
        A valid output type of ``rst`` or ``github``

    Returns
    -------
    Any
        The report.
    """
    if output_type not in VALID_OUTPUTS:
        raise ValueError(f"Invalid output_type provided: {output_type}")

    headers = [
        "Environment",
        "Setup successful",
        "Passing tests",
        "Upgraded packages",
        "Lowered packages",
        "Package version",
    ]
    rows: List[List] = []
    for env in testers:
        upgraded = env.upgraded_packages()
        lowered = env.lowered_packages()
        for pkg in upgraded:
            rows.append(
                [
                    env.envname,
                    env.setup_status,
                    env.status,
                    pkg["name"],
                    "",
                    pkg["version"],
                ]
            )
        for pkg in lowered:
            rows.append(
                [
                    env.envname,
                    env.setup_status,
                    env.status,
                    "",
                    pkg["name"],
                    pkg["version"],
                ]
            )

    return tabulate(rows, headers=headers, tablefmt=output_type)



"""Test modifying and adjusting the schema."""

from edgetest.schema import EdgetestValidator, Schema


def test_add_envoption():
    """Test adding an environment option."""
    schema = Schema()
    schema.add_envoption(
        "mylist", {"type": "list", "schema": {"schema": {"type": "string"}}}
    )

    conf = {"envs": [{"name": "myenv", "upgrade": "\nmypackage", "mylist": ["heyo"]}]}

    validator = EdgetestValidator(schema=schema.schema)

    assert validator.validate(conf)

    conf = {
        "envs": [{"name": "myenv", "upgrade": "\nmypackage", "mylist": "not a list"}]
    }

    assert not validator.validate(conf)


def test_add_global_option():
    """Test adding a global option to the schema."""
    schema = Schema()
    schema.add_globaloption("title", {"type": "string"})

    conf = {"title": "My title", "envs": [{"name": "myenv", "upgrade": "\nmyupgrade"}]}

    validator = EdgetestValidator(schema=schema.schema)

    assert validator.validate(conf)

    conf = {"title": 3, "envs": [{"name": "myenv", "upgrade": "\nmyupgrade"}]}

    assert not validator.validate(conf)



"""Create a fake package and test."""
import configparser
import sys
from pathlib import Path

import pytest
from click.testing import CliRunner

from edgetest.interface import cli

SETUP_CFG = """
[metadata]
name = toy_package
version = 0.1.0
description = Fake description
python_requires =
    >=3.7.0

[options]
zip_safe = False
include_package_data = True
packages = find:
install_requires =
    pandas<=1.2.0

[options.extras_require]
tests =
    pytest

[edgetest]
extras =
    tests
"""

SETUP_CFG_LOWER = """
[metadata]
name = toy_package
version = 0.1.0
description = Fake description
python_requires =
    >=3.7.0

[options]
zip_safe = False
include_package_data = True
packages = find:
install_requires =
    pandas<=1.2.0,>=1.0.0

[options.extras_require]
tests =
    pytest

[edgetest]
extras =
    tests

[edgetest.envs.lower_env]
lower =
    pandas
"""


SETUP_CFG_DASK = """
[metadata]
name = toy_package
version = 0.1.0
description = Fake description
python_requires =
    >=3.7.0

[options]
zip_safe = False
include_package_data = True
packages = find:
install_requires =
    scikit-learn>=1.0,<=1.2.0
    dask[dataframe]<=2022.1.0,>=2021.6.1

[options.extras_require]
tests =
    pytest

[edgetest.envs.core]
extras =
    tests
upgrade =
    Scikit_Learn
    Dask[DataFrame]

[edgetest.envs.lower_env]
extras =
    tests
lower =
    scikit_Learn
    Dask[DataFrame]
"""


SETUP_PY = """
from setuptools import setup

setup()
"""

MODULE_CODE = """
def main() -> str:
    print("Hello world")
"""

TEST_CODE = """
from toy_package import main

def test_func():
    main()
"""

PY_VER = f"python{sys.version_info.major}.{sys.version_info.minor}"


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
@pytest.mark.integration
def test_toy_package():
    """Test using edgetest with a toy package."""
    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("setup.cfg", "w") as outfile:
            outfile.write(SETUP_CFG)
        with open("setup.py", "w") as outfile:
            outfile.write(SETUP_PY)
        # Make a directory for the module
        Path(loc, "toy_package").mkdir()
        with open(Path(loc, "toy_package", "__init__.py"), "w") as outfile:
            outfile.write(MODULE_CODE)
        # Make a directory for the tests
        Path(loc, "tests").mkdir()
        with open(Path(loc, "tests", "test_main.py"), "w") as outfile:
            outfile.write(TEST_CODE)

        # Run the CLI
        result = runner.invoke(cli, ["--config=setup.cfg"])

        assert result.exit_code == 0
        assert Path(loc, ".edgetest").is_dir()
        assert Path(loc, ".edgetest", "pandas").is_dir()
        assert "pandas" in result.stdout

        if not sys.platform == "win32":
            assert Path(
                loc, ".edgetest", "pandas", "lib", PY_VER, "site-packages", "pandas"
            ).is_dir()


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
@pytest.mark.integration
def test_toy_package_lower():
    """Test lower bounds with a toy package."""
    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("setup.cfg", "w") as outfile:
            outfile.write(SETUP_CFG_LOWER)
        with open("setup.py", "w") as outfile:
            outfile.write(SETUP_PY)
        # Make a directory for the module
        Path(loc, "toy_package").mkdir()
        with open(Path(loc, "toy_package", "__init__.py"), "w") as outfile:
            outfile.write(MODULE_CODE)
        # Make a directory for the tests
        Path(loc, "tests").mkdir()
        with open(Path(loc, "tests", "test_main.py"), "w") as outfile:
            outfile.write(TEST_CODE)

        # Run the CLI
        result = runner.invoke(cli, ["--config=setup.cfg"])

        assert result.exit_code == 0
        assert Path(loc, ".edgetest").is_dir()
        assert Path(loc, ".edgetest", "lower_env").is_dir()
        assert "pandas" in result.stdout

        if not sys.platform == "win32":
            assert Path(
                loc, ".edgetest", "lower_env", "lib", PY_VER, "site-packages", "pandas"
            ).is_dir()


@pytest.mark.integration
def test_toy_package_dask():
    """Test using edgetest with a toy package."""
    runner = CliRunner()

    with runner.isolated_filesystem() as loc:
        with open("setup.cfg", "w") as outfile:
            outfile.write(SETUP_CFG_DASK)
        with open("setup.py", "w") as outfile:
            outfile.write(SETUP_PY)
        # Make a directory for the module
        Path(loc, "toy_package").mkdir()
        with open(Path(loc, "toy_package", "__init__.py"), "w") as outfile:
            outfile.write(MODULE_CODE)
        # Make a directory for the tests
        Path(loc, "tests").mkdir()
        with open(Path(loc, "tests", "test_main.py"), "w") as outfile:
            outfile.write(TEST_CODE)

        # Run the CLI
        result = runner.invoke(cli, ["--config=setup.cfg", "--export"])

        assert result.exit_code == 0
        assert Path(loc, ".edgetest").is_dir()

        for envname in ("core", "lower_env"):
            assert Path(loc, ".edgetest", envname).is_dir()

            if not sys.platform == "win32":
                assert Path(
                    loc, ".edgetest", envname, "lib", PY_VER, "site-packages", "dask"
                ).is_dir()
                assert Path(
                    loc, ".edgetest", envname, "lib", PY_VER, "site-packages", "pandas"
                ).is_dir()
                assert Path(
                    loc, ".edgetest", envname, "lib", PY_VER, "site-packages", "sklearn"
                ).is_dir()

        config = configparser.ConfigParser()
        config.read("setup.cfg")

        assert "scikit-learn" in config["options"]["install_requires"]
        assert "dask[dataframe]" in config["options"]["install_requires"]
        assert config["edgetest.envs.core"]["extras"] == "\ntests"
        assert (
            config["edgetest.envs.core"]["upgrade"] == "\nScikit_Learn\nDask[DataFrame]"
        )
        assert config["edgetest.envs.lower_env"]["extras"] == "\ntests"
        assert (
            config["edgetest.envs.lower_env"]["lower"]
            == "\nscikit_Learn\nDask[DataFrame]"
        )
        assert "dask" in result.stdout
        assert "scikit-learn" in result.stdout
